; _______________________________________________________________________________________
;|                                                                                       |
;| ..::FreshLib::..  Free, open source. Licensed under "BSD 2-clause" license."          |
;|_______________________________________________________________________________________|
;
;  Description: Memory based images manipulation library.
;
;  Target OS: Win32
;
;  Dependencies: memory.asm
;
;  Notes:
;_________________________________________________________________________________________


; only 16, 24 and 32 bpp are supported.
; It is stupid to support palleted modes and even 24bpp is not good...
proc CreateImage, .width, .height, .bpp
.bmi BITMAPINFOHEADER
begin
        push    ebx ecx edx esi

        stdcall GetMem, sizeof.TImage
        jc      .finish

        mov     esi, eax

        cmp     [.width], 0
        jge     @f
        mov     [.width], 0
@@:
        cmp     [.height], 0
        jge     @f
        mov     [.height], 0
@@:
        push    [.width] [.height] [.bpp]
        pop     [esi+TImage.bpp] [esi+TImage.height] [esi+TImage.width]

        lea     eax, [.bmi]
        mov     [eax+BITMAPINFOHEADER.biSize], sizeof.BITMAPINFOHEADER
        mov     [eax+BITMAPINFOHEADER.biPlanes], 1
        mov     [eax+BITMAPINFOHEADER.biCompression], BI_RGB
        mov     [eax+BITMAPINFOHEADER.biSizeImage], 0
        mov     [eax+BITMAPINFOHEADER.biXPelsPerMeter], 96      ; ???
        mov     [eax+BITMAPINFOHEADER.biYPelsPerMeter], 96      ; ???
        mov     [eax+BITMAPINFOHEADER.biClrUsed], 0
        mov     [eax+BITMAPINFOHEADER.biClrImportant], 0

        mov     ecx, [.bpp]
        mov     [eax+BITMAPINFOHEADER.biBitCount], cx
        push    [.width] [.height]
        neg     dword [esp]
        pop     [eax+BITMAPINFOHEADER.biHeight] [eax+BITMAPINFOHEADER.biWidth]

        lea     ecx, [esi+TImage.pData]
        invoke  CreateDIBSection, 0, eax, DIB_RGB_COLORS, ecx, 0, 0
        mov     [esi+TImage.raster], eax

        clc
        mov     eax, esi

.finish:
        pop     esi edx ecx ebx
        return
endp


proc DestroyImage, .ptrImage
begin
        push    eax ecx edx esi

        mov     esi, [.ptrImage]
        cmp     esi, 0
        je      @f
        invoke  DeleteObject, [esi+TImage.raster]
        stdcall FreeMem, esi
@@:
        pop     esi edx ecx eax
        return
endp




proc DrawImage, .context, .pImage, .x, .y
begin
        push    eax ecx edx esi edi
        mov     esi, [.pImage]
        test    esi, esi
        jz      .exit

        stdcall AllocateContext, [esi+TImage.raster]
        jc      .exit
        mov     edi, eax

        mov     ecx, [.context]
        mov     eax, [ecx+TContext.drawmode]
        dec     eax
        and     eax, $0f
        invoke  BitBlt, [ecx+TContext.handle], [.x], [.y], [esi+TImage.width], [esi+TImage.height], [edi+TContext.handle], 0, 0, [4*eax+__RopTable]

        stdcall ReleaseContext, edi

.exit:
        pop     edi esi edx ecx eax
        return
endp

; NEEDS some edit to fill the zeroes.

if used __RopTable
__RopTable dd  BLACKNESS, PATPAINT,    0,          NOTSRCCOPY, \
               0,         DSTINVERT,   SRCINVERT,  SRCERASE,   \
               SRCAND,    0,           0,          0,          \
               SRCCOPY,   MERGEPAINT,  SRCPAINT,   WHITENESS
end if


