; _______________________________________________________________________________________
;|                                                                                       |
;| ..::FreshLib::..  Free, open source. Licensed under "BSD 2-clause" license."          |
;|_______________________________________________________________________________________|
;
;  Description: Objects template engine
;
;  Target OS: Any
;
;  Dependencies: objects.asm
;
;  Notes:
;_________________________________________________________________________________________
module "Templates library"

tfChild  = 0
tfParent = 1
tfEnd    = 2


struct TObjTemplate
  .flags     dd  ?         ; tfParent; tfEnd
  .class     dd  ?
  .ptrVar    dd  ?         ; can be NULL
  .paramsize dd  ?
  .params:
ends


macro ObjTemplate  flags, class, name, [id, param] {
common
local ..paramsize, ..prms
      dd  flags
      dd  C#class
      dd  name
      dd  ..paramsize
..prms:
forward
local ..value
      dd  T#class#.#id
      dd  ..value
common
      dd  -1
forward
  if param eqtype 2
    ..value = param
  else
    ..value db param, 0
  end if
common
  name dd ?
  ..paramsize = $ - ..prms
}


; returns the last created window in ebx.

proc CreateFromTemplate, .ptrTemplate, .parent
begin
        push    esi

        mov     esi, [.ptrTemplate]
        stdcall _DoCreateFromTemplate, [.parent]

        pop     esi
        return
endp


proc _DoCreateFromTemplate, .parent
begin
        push    eax ecx edx

.createloop:
        stdcall Create, [esi+TObjTemplate.class]
        execute [.parent], TObject.AddChild, ebx

        stdcall IsObject, ebx, CWindow
        jc      .border_ok

        cmp     [.parent], 0
        jne     .border_ok

        stdcall Set, ebx, TWindow.borderKind, borderFull

.border_ok:
        lea     ecx, [esi+TObjTemplate.params]

.paramloop:
        cmp     dword [ecx], -1
        je      .paramsok

        push    ebx ecx edx esi edi
        stdcall Set, ebx, [ecx], [ecx+4]
        pop     edi esi edx ecx ebx
        add     ecx, 8
        jmp     .paramloop

.paramsok:
        mov     eax, [esi+TObjTemplate.ptrVar]
        test    eax, eax
        jz      .oncreate

        mov     [ebx+TObject.ptrVar], eax       ; back link to the object instance variable.
        mov     [eax], ebx      ; returned pointer/handle to the object.

.oncreate:
        cmp     [ebx+TObject.OnCreate], 0
        je      .gonext

; execute user OnCreate event.
        pushad
        stdcall [ebx+TObject.OnCreate], ebx
        popad

.gonext:
        mov     edx, [esi+TObjTemplate.flags]

        mov     eax, sizeof.TObjTemplate
        add     eax, [esi+TObjTemplate.paramsize]
        add     esi, eax

        test    edx, tfParent
        jz      .childrenok

        push    ebx
        stdcall _DoCreateFromTemplate, ebx
        pop     ebx

.childrenok:
        test    edx, tfEnd
        jz      .createloop

.exit:
        pop     edx ecx eax
        return
endp


endmodule