; _______________________________________________________________________________________
;|                                                                                       |
;| ..::FreshLib::..  Free, open source. Licensed under "BSD 2-clause" license."          |
;|_______________________________________________________________________________________|
;
;  Description: TImageLabel object class
;
;  Target OS: Any
;
;  Dependencies:
;
;  Notes: Represents GUI picture or image.
;_________________________________________________________________________________________
module "TImageLabel library"

ObjectClass ImageLabel,                 \
            Window,                     \
            0,                          \
            0,                          \
            TImageLabel.Get,            \
            TImageLabel.Set,                \
            0,                          \
            TImageLabel.SysEventHandler

iaLeft = 0
iaRight = 1
iaCenter = 2
iaStretchH = 3

iaTop = $0
iaBottom = $100
iaMiddle = $200
iaStrechV = $300

object TImageLabel, TWindow
  .FImageAlign dd ?
  .FImage      dd ?
  .FMask       dd ?
  .FBackground dd ?

  param .ImageAlign       ; image align flags.
  param .Image            ; TImage object
  param .Mask
endobj



proc TImageLabel.Get, .obj, .paramID
begin
        mov     eax, [.obj]

        stdcall IsObject, eax, CImageLabel
        jc      .exit

        cmp     [.paramID], TImageLabel.ImageAlign
        je      .getimagealign

        cmp     [.paramID], TImageLabel.Image
        je      .getimage

        cmp     [.paramID], TImageLabel.Mask
        je      .getmask

        stc
.exit:
        return

.getimagealign:
        mov     eax, [eax+TImageLabel.FImageAlign]
        clc
        return

.getimage:
        mov     eax, [eax+TImageLabel.FImage]
        clc
        return

.getmask:
        mov     eax, [eax+TImageLabel.FMask]
        clc
        return
endp



proc TImageLabel.Set, .obj, .paramID, .value
begin
        push    eax

        mov     eax, [.obj]
        stdcall IsObject, eax, CImageLabel
        jc      .exit

        cmp     [.paramID], TImageLabel.ImageAlign
        je      .setimagealign

        cmp     [.paramID], TImageLabel.Image
        je      .setimage

        cmp     [.paramID], TImageLabel.Mask
        je      .setmask

        stc
.exit:
        pop     eax
        return

.setimage:
        push    [.value]
        pop     [eax+TImageLabel.FImage]
        jmp     .refresh

.setmask:
        push    [.value]
        pop     [eax+TImageLabel.FMask]
        jmp     .refresh

.setimagealign:
        push    [.value]
        pop     [eax+TImageLabel.FImageAlign]

.refresh:
        execute [.obj], TWindow.Refresh
        clc
        pop     eax
        return
endp



proc TImageLabel.SysEventHandler, .pObj, .pEvent
begin
        push    eax ebx ecx edx esi

        mov     esi, [.pObj]
        mov     ebx, [.pEvent]
        mov     eax, [ebx+TSysEvent.event]

        cmp     eax, sePaint
        je      .onpaint

.finish:
        stc
        pop     esi edx ecx ebx eax
        return

.onpaint:
  locals
   .bounds TBounds
  endl
        mov     ecx, [esi+TImageLabel._width]
        mov     eax, [esi+TImageLabel._height]
        mov     [.bounds.x], 0
        mov     [.bounds.y], 0
        mov     [.bounds.width], ecx
        mov     [.bounds.height], eax

        stdcall SetDrawMode, [ebx+TPaintEvent.context], cmCopy
        stdcall DrawFillRect, [ebx+TPaintEvent.context], [.bounds.x], [.bounds.y], [.bounds.width], [.bounds.height], [clDialogBk]

        mov     eax, [esi+TImageLabel.FImage]
        mov     ecx, [.bounds.width]
        mov     edx, [.bounds.height]
        sub     ecx, [eax+TImage.width]
        sub     edx, [eax+TImage.height]

        mov     eax, [esi+TImageLabel.FImageAlign]

        cmp     al, iaRight
        je      .x_ok
        cmp     al, iaCenter
        je      .x_center
        xor     ecx, ecx
        jmp     .x_ok
.x_center:
        sar     ecx, 1
.x_ok:
        cmp     ah, iaBottom /256
        je      .y_ok
        cmp     ah, iaMiddle / 256
        je      .y_center
        xor     edx, edx
        jmp     .y_ok
.y_center:
        sar     edx, 1
.y_ok:
        stdcall DrawMaskedImage, [ebx+TPaintEvent.context], [esi+TImageLabel.FMask], [esi+TImageLabel.FImage], ecx, edx
        jmp     .finish
endp








endmodule