; _______________________________________________________________________________________
;|                                                                                       |
;| ..::FreshLib::..  Free, open source. Licensed under "BSD 2-clause" license."          |
;|_______________________________________________________________________________________|
;
;  Description: TObject object class.
;
;  Target OS: Any
;
;  Dependencies:
;
;  Notes: TObject is the root member of FreshLib objects tree.
;_________________________________________________________________________________________
module "TObject library"

ObjectClass  Object,                    \
             Root,                      \
             0,                         \
             TObject.Destroy,           \
             0,                         \
             0,                         \
             0,                         \
             0


object TObject
  .ptrClass     dd ?    ; pointer to the object class structure.

  if options.Threads
    .mxAccess     TMutex  ; access mutex. Procedures from objects.asm will use it to ensure thread safety.
  end if


  .ptrVar       dd ?    ; pointer to the variable that keep the pointer to the object. This field can be 0. If not, Destroy procedure will set it to 0 on destroy.

  .OnCreate     dd ?    ; user event handlers.
  .OnDestroy    dd ?

  method .AddChild, .objchild    ; this method is not implemented in TObject, but every descendent can implement it's own parent/child system.
endobj



proc TObject.Destroy, .obj
begin
        push    eax
        mov     eax, [.obj]

        cmp     [eax+TObject.OnDestroy], 0
        je      .exit

        pushad
        stdcall [eax+TObject.OnDestroy], eax
        popad

.exit:
        pop     eax
        return
endp



endmodule