; _______________________________________________________________________________________
;|                                                                                       |
;| ..::FreshLib::..  Free, open source. Licensed under "BSD 2-clause" license."          |
;|_______________________________________________________________________________________|
;
;  Description: OS dependent part of the network library.
;
;  Target OS: Linux
;
;  Dependencies:
;
;  Notes:
;_________________________________________________________________________________________


serrAccess              = EACCES
serrNoMemory            = ENOMEM
serrProtocolNotSupported= EPROTONOSUPPORT

serrAddessInUse         = EADDRINUSE
serrNotASocket          = ENOTSOCK

serrWouldBlock          = EWOULDBLOCK
serrBadSocket           = EBADF
serrConnRefused         = ECONNREFUSED
serrConnReset           = ECONNRESET
serrDestAddrReq         = EDESTADDRREQ
serrInvalidAddress      = EFAULT
serrInterrupt           = EINTR
serrInvalidArgument     = EINVAL
serrIsConnected         = EISCONN



proc __SocketError
begin
        test    eax, eax
        jns     .no_error
        neg     eax
        stc
        return

.no_error:
        clc
        return
endp


proc SocketCreate, .protocol_family, .socket_type, .protocol
.handle dd ?
.level  dd ?
.option dd ?
.pParam dd ?
.Size   dd ?
.buff   rd 4
begin
        push    ebx ecx edx

        mov     eax, sys_socketcall
        mov     ebx, SYS_SOCKET
        lea     ecx, [.protocol_family]
        int     $80

        call    __SocketError
        jc      .exit

        mov     [.handle], eax
        mov     [.level], SOL_SOCKET
        mov     [.option], SO_REUSEADDR
        lea     eax, [.buff]
        mov     [.pParam], eax
        mov     [.buff], TRUE
        mov     [.Size], 4

        mov     eax, sys_socketcall
        mov     ebx, SYS_SETSOCKOPT
        lea     ecx, [.handle]
        int     $80

        mov     eax, [.handle]
        clc

.exit:
        pop     edx ecx ebx
        return
endp



proc SocketClose, .hSocket
;.buffer rb 1024
;.params rd 2
begin
        push    ebx ecx edx

;        push    [.hSocket]
;        pop     [.params]
;        mov     [.params+4], 1
;
;        mov     eax, sys_socketcall
;        mov     ebx, SYS_SHUTDOWN
;        lea     ecx, [.params]
;        int     $80
;        call    __SocketError
;        jc      .finish
;
;.loop:
;        lea     ecx, [.buffer]
;        stdcall SocketReceive, [.hSocket], ecx, 1024, 0
;        jc      .finish
;        test    eax, eax
;        jnz     .loop
;
;.finish:
        mov     eax, sys_close
        mov     ebx, [.hSocket]
        int     $80
        call    __SocketError

        pop     edx ecx ebx
        return
endp



proc SocketConnect, .hSocket, .pAddress
.sock dd ?
.addr dd ?
.size dd ?
begin
        push    ebx ecx edx

        push    [.hSocket] [.pAddress]
        pop     [.addr] [.sock]
        mov     [.size], sizeof.TSocketAddress

        mov     eax, sys_socketcall
        mov     ebx, SYS_CONNECT
        lea     ecx, [.sock]
        int     $80

        call    __SocketError
        pop     edx ecx ebx
        return
endp



proc SocketBind, .hSocket, .pAddressIn
.sock dd ?
.addr dd ?
.size dd ?
begin
        push    ebx ecx edx

        push    [.hSocket] [.pAddressIn]
        pop     [.addr] [.sock]
        mov     [.size], sizeof.TSocketAddressIn

        mov     eax, sys_socketcall
        mov     ebx, SYS_BIND
        lea     ecx, [.sock]
        int     $80

        call    __SocketError
        pop     edx ecx ebx
        return
endp



proc SocketListen, .hSocket, .maxPending
begin
        push    ebx ecx edx

        mov     eax, sys_socketcall
        mov     ebx, SYS_LISTEN
        lea     ecx, [.hSocket]
        int     $80

        call    __SocketError
        pop     edx ecx ebx
        return
endp



proc SocketAccept, .hSocket, .pAddress
.sock    dd ?
.addr    dd ?
.addrlen dd ?
begin
        push    ebx edx

        push    [.hSocket] [.pAddress]
        pop     [.addr] [.sock]
        mov     [.addrlen], sizeof.TSocketAddress

        cmp     [.addr], 0
        jne     @f
        mov     [.addrlen], 0
@@:
        mov     eax, sys_socketcall
        mov     ebx, SYS_ACCEPT
        lea     ecx, [.sock]
        int     $80

        call    __SocketError
        mov     ecx, [.addrlen]

        pop     edx ebx
        return
endp



proc SocketSend, .hSocket, .pBuffer, .DataLen, .flags
begin
        push    ebx ecx edx

        or      [.flags], MSG_NOSIGNAL
        mov     eax, sys_socketcall
        mov     ebx, SYS_SEND
        lea     ecx, [.hSocket]
        int     $80
        call    __SocketError
        pop     edx ecx ebx
        return
endp


proc SocketReceive, .hSocket, .pBuffer, .BufferSize, .flags
begin
        push    ebx ecx edx
        mov     eax, sys_socketcall
        mov     ebx, SYS_RECV
        lea     ecx, [.hSocket]
        int     $80
        call    __SocketError
        pop     edx ecx ebx
        return
endp


proc SocketSendTo, .hSocket, .pBuffer, .DataLen, .flags, .pAddressTo
.sock  dd ?
.buff  dd ?
.len   dd ?
.flgs  dd ?
.addr  dd ?
.alen  dd ?
begin
        push    ebx ecx edx

        lea     ecx, [.sock]
        mov     edx, 16

.copy:
        mov     eax, [.hSocket+4*edx]
        mov     [ecx+4*edx], eax
        dec     edx
        jns     .copy
        mov     [.alen], sizeof.TSocketAddress

        mov     eax, sys_socketcall
        mov     ebx, SYS_SENDTO
        int     $80
        call    __SocketError

        pop     edx ecx ebx
        return
endp


proc SocketReceiveFrom, .hSocket, .pBuffer, .BufferSize, .flags, .pAddressFrom
.sock  dd ?
.buff  dd ?
.len   dd ?
.flgs  dd ?
.addr  dd ?
.alen  dd ?
begin
        push    ebx edx

        lea     ecx, [.sock]
        mov     edx, 16

.copy:
        mov     eax, [.hSocket+4*edx]
        mov     [ecx+4*edx], eax
        dec     edx
        jns     .copy
        mov     [.alen], sizeof.TSocketAddress

        mov     eax, sys_socketcall
        mov     ebx, SYS_RECVFROM
        int     $80
        call    __SocketError
        mov     ecx, [.alen]
        pop     edx ebx
        return
endp