uglobal
  hProjManager    dd 0  ; handle of project manager
  ptrProjectFile  dd 0  ; pointer to TOpenFile with ftProject, currently open in the project manager.
  fProjModified   dd 0  ; flag "modified" for current open project.

  poMemory       dd 0  ; The memory for the compiler
  poBinNameAuto  dd 0  ; The flag whether the binary name will be auto created of fixed
  poBinName      dd 0  ; Handle to the binary name.
  poVarArray     dd 0  ; points to the array with environment variables for the project.

  imlCategories   dd 0  ; Image list for tree-view control

  ptrMRUProj      dd 0  ; pointer to MRU list for projects.

  hTreePopup      dd 0
  hListPopup      dd 0
  hSettingsPopup  dd 0
  hSettingsString dd 0
endg


cProjTitle       text 'Project manager'
cSettingsCaption text 'Settings: '


iglobal
  CoolMenu mnuProjTree,           \
    mfNormal, actAddCategory,     \
    mfNormal, actEditCategory,    \
    mfSeparator, NONE,            \
    mfNormal, actMoveCategoryUp,  \
    mfNormal, actMoveCategoryDn

  CoolMenu mnuProjList,         \
    mfNormal, actAddFile

  CoolMenu mnuSettings,         \
    mfSeparator, NONE,          \
    mfNormal, actProjectOptions

endg




;----------------------------------------
; Creates project manager common objects.
; Call this only once.
;----------------------------------------
proc InitProjectManager
begin
; Read project MRU list
        stdcall LoadMRUFromIni, 16, [hIniFileName], cMRUProjSection
        mov     [ptrMRUProj], eax

        stdcall ImageList_LoadGif, [hInstance], resBmpProjMan, 16, FALSE
        mov     [imlCategories], eax

        stdcall CreateCoolMenu, mnuProjTree, TRUE
        mov     [hTreePopup], eax

        stdcall CreateCoolMenu, mnuProjList, TRUE
        mov     [hListPopup], eax

        stdcall CreateCoolMenu, mnuSettings, TRUE
        mov     [hSettingsPopup], eax

        mov     [poBinNameAuto], TRUE
        mov     [poMemory], 65536

        stdcall StrNew
        mov     [poBinName], eax

        return
endp


proc FinalizeProjectManager
begin
        cmp     [ptrProjectFile], NULL
        je      .closeproj

        stdcall CloseFile, [ptrProjectFile]

.closeproj:
        stdcall SaveMRUToIni, [ptrMRUProj], [hIniFileName], cMRUProjSection
        stdcall DestroyMRUList, [ptrMRUProj]
        invoke  ImageList_Destroy, [imlCategories]

        invoke  DestroyMenu, [hTreePopup]
        invoke  DestroyMenu, [hListPopup]
        invoke  DestroyMenu, [hSettingsPopup]
        return
endp


proc CreateProjectManager, .ptrOpenFile, .hTemplate
begin
        push    ebx esi
        xor     ebx, ebx

        mov     esi, [.ptrOpenFile]

        cmp     [ptrProjectFile], ebx
        je      .onlyone

        stdcall CloseFile, [ptrProjectFile]
        cmp     eax, cfrDestroyed
        jne     .fault

.onlyone:
        mov     [ptrProjectFile], esi

; Initial project options... they could be changed after CEM_LOADFILE
        mov     [poMemory], 65536
        mov     [poBinNameAuto], TRUE

        cmp     [poBinName], 0
        je      @f
        stdcall StrDel, [poBinName]
@@:
        stdcall StrNew
        mov     [poBinName], eax

        invoke   CreateWindowExA, WS_EX_TOOLWINDOW, cFormClassName, cProjTitle,  \
                                 WS_VISIBLE or WS_CAPTION or WS_CLIPCHILDREN or \
                                 WS_SIZEBOX or WS_SYSMENU,                      \
                                 ebx, ebx , 160, 400,                           \
                                 [hApplication], ebx, [hInstance], ebx

        mov     [hProjManager], eax
        mov     [esi+TOpenFile.hEditor], eax
        mov     [esi+TOpenFile.projectID], ebx
        stdcall SubclassWindow, eax, ProjManProc

        invoke  SendMessageA, [hProjManager], CEM_LOADFILE, [.hTemplate], ebx
        push    eax
        stdcall AutoArrangeWindows
        pop     eax

.finish:
        pop     esi ebx
        return

.fault:
        mov     eax, errCanceled
        pop     esi ebx
        return
endp



VarMenuBaseID = 20

proc CreateParametersMenu
.mii MENUITEMINFO
.pos dd ?
.pos_in_list dd ?
.buffer rb 256
begin
        push    ebx ecx esi

        xor     eax, eax
        cmp     [hSettingsString], eax
        jne     @f
        stdcall StrDel, [hSettingsString]
@@:
        stdcall StrDup, cSettingsCaption
        mov     [hSettingsString], eax

; first clear to the first separator.
.deleteloop:
        lea     ebx, [.mii]
        mov     [ebx+MENUITEMINFO.cbSize], sizeof.MENUITEMINFO
        mov     [ebx+MENUITEMINFO.fMask], MIIM_TYPE
        mov     [ebx+MENUITEMINFO.hSubMenu], 0
        lea     eax, [.buffer]
        mov     [ebx+MENUITEMINFO.dwTypeData], eax
        mov     [ebx+MENUITEMINFO.cch], 256
        invoke  GetMenuItemInfoA, [hSettingsPopup], 0, TRUE, ebx
        test    [ebx+MENUITEMINFO.fType], MFT_SEPARATOR
        jnz     .insertvalues

        invoke  DeleteMenu, [hSettingsPopup], 0, MF_BYPOSITION
        jmp     .deleteloop

; then check for multivalue elements in poVarArray
.insertvalues:
        mov     [.pos], 0
        mov     [.pos_in_list], VarMenuBaseID

        mov     esi, [poVarArray]
        test    esi, esi
        jz      .finish

        mov     ecx, [esi+TArray.count]
        lea     esi, [esi+TArray.array]

        cmp     ecx, 0
        je      .finish

.arrayloop:
        stdcall StrCharPos, [esi+TEnvVarType.pValue], '|'       ; check for separator.
        test    eax, eax
        jz      .next_item

        push    ecx
        mov     ecx, eax

        stdcall CreatePopupFromString, [esi+TEnvVarType.pValue], [.pos_in_list]
        mov     [ebx+MENUITEMINFO.hSubMenu], eax

; create menu string.
        mov     byte [ecx], 0
        stdcall StrDup, [esi+TEnvVarType.pName]
        push    eax

        stdcall StrCharCat, eax, 09
        stdcall StrCat, eax, [esi+TEnvVarType.pValue]
        stdcall StrCat, [hSettingsString], [esi+TEnvVarType.pValue]
        stdcall StrCharCat, [hSettingsString], ','
        stdcall StrCharCat, [hSettingsString], ' '
        mov     byte [ecx], '|'

; insert menu item

        mov     [ebx+MENUITEMINFO.cbSize], sizeof.MENUITEMINFO
        mov     [ebx+MENUITEMINFO.fMask], MIIM_TYPE or MIIM_SUBMENU
        mov     [ebx+MENUITEMINFO.fType], MFT_STRING
        stdcall StrPtr, eax
        mov     [ebx+MENUITEMINFO.dwTypeData], eax
        mov     [ebx+MENUITEMINFO.cch], 0
        invoke  InsertMenuItemA, [hSettingsPopup], [.pos], TRUE, ebx
        inc     [.pos]

        stdcall StrDel ; from the stack.

        pop     ecx

.next_item:
        add     esi, sizeof.TEnvVarType
        inc     [.pos_in_list]
        dec     ecx
        jnz     .arrayloop

.finish:
        invoke  GetDlgItem, [hProjManager], ProjSettingsID
        invoke  RedrawWindow, eax, 0, 0, RDW_INVALIDATE

        pop     esi ecx ebx
        return
endp



proc CreatePopupFromString, .hString, .baseid
.mii MENUITEMINFO
.pos dd ?
begin
        push    ebx ecx esi edi

        invoke  CreatePopupMenu
        mov     edi, eax

        stdcall StrPtr, [.hString]
        mov     esi, eax
        lea     ebx, [.mii]

        mov     [.pos], 0

.stringsplit:
        mov     eax, [.baseid]
        shl     eax, 8
        add     eax, [.pos]

        mov     [ebx+MENUITEMINFO.wID], eax
        mov     [ebx+MENUITEMINFO.cbSize], sizeof.MENUITEMINFO
        mov     [ebx+MENUITEMINFO.fMask], MIIM_TYPE or MIIM_ID
        mov     [ebx+MENUITEMINFO.fType], MFT_STRING
        mov     [ebx+MENUITEMINFO.dwTypeData], esi
        mov     [ebx+MENUITEMINFO.cch], 0
        mov     [ebx+MENUITEMINFO.hSubMenu], 0

        xor     eax, eax
.searchloop:
        lodsb
        cmp     al, '|'
        je      .splithere
        cmp     al, 0
        jne      .searchloop

.splithere:
        push    eax
        mov     byte [esi-1], 0

        invoke  InsertMenuItemA, edi, 0, TRUE, ebx
        inc     [.pos]

        pop     eax
        mov     [esi-1], al
        cmp     eax, 0
        jne     .stringsplit

        mov     eax, edi
        pop     edi esi ecx ebx
        return
endp





msgPUC1 text 'The file: "'
msgPUC2 text '" is part of saved project, but it doesn''t exists on the disk.', 13, 10, \
               'Do you want to add it to the project anyway?'
msgPUCtitle text  'Error adding file to the project.'


proc ProjectInconsistent, .ptrOpenFile
begin
        push    ebx
        mov     ebx, [.ptrOpenFile]

        stdcall StrDup, msgPUC1
        stdcall StrCat, eax, [ebx+TOpenFile.hFileName]
        stdcall StrCat, eax, msgPUC2
        push    eax
        stdcall StrPtr, eax

        invoke  MessageBoxA, [hApplication], eax, msgPUCtitle, MB_YESNO or MB_ICONWARNING
        stdcall StrDel ; from the stack
        pop     ebx
        return
endp




;-----------------------------------------------------
; Dialog proc for Project manager.
;-----------------------------------------------------
winproc ProjManProc

.tvi    TVINSERTSTRUCT
.lvi    LVITEM
.col    LVCOLUMN
.count  dd ?
.str    rb 1024
.str2   rb 256
.ptr    dd ?
.pnt    POINT

begin

ondefault
        stc
        return


onmessage PMM_FILECHANGED

        stdcall SearchForFile, [.hwnd], [.wparam]
        cmp     eax, -1
        je      .endchanged

        push    eax

        invoke  SendMessageA, [.hwnd], PMM_REMOVEFILE, [.wparam], 0
        invoke  SendMessageA, [.hwnd], PMM_ADDFILE, [.wparam] ; lparam from the stack

.endchanged:
        clc
        return

onmessage PMM_ADDFILE

        mov     [fProjModified], TRUE
        mov     eax, [.wparam]
        cmp     [eax+TOpenFile.projectID], FALSE
        je      .projok

        invoke  SendMessageA, [.hwnd], PMM_REMOVEFILE, eax, 0

.projok:
        cmp     [.lparam], 0
        jne     .addtocategory

; Add as main file.
        mov     eax, [.wparam]
        cmp     [eax+TOpenFile.ptrType], ftMainSource
        je      .typeok
        cmp     [eax+TOpenFile.ptrType], ftCommonSource
        jne     .erroradd
        mov     [eax+TOpenFile.ptrType], ftMainSource   ; change file type.
.typeok:
        mov     [eax+TOpenFile.projectID], TRUE

        stdcall AddToListView, [.hwnd], ProjMainCtrlID, [.wparam]
        invoke  SendDlgItemMessageA, [.hwnd], ProjMainCtrlID, LVM_SETCOLUMNWIDTH, 0, LVSCW_AUTOSIZE

        stdcall FileChanged, [.wparam], fcEditor
        jmp     .endadd

.addtocategory:
        mov     eax, [.wparam]
        cmp     [eax+TOpenFile.ptrType], ftMainSource
        jne     .typeok2

        mov     [eax+TOpenFile.ptrType], ftCommonSource

.typeok2:
        cmp     [eax+TOpenFile.projectID], FALSE
        jne     .erroradd

        mov     eax, [.lparam]
        cmp     eax, -1
        jne     .addit

.getcurrent:
        invoke  SendDlgItemMessageA, [.hwnd], ProjTreeCtrlID, TVM_GETNEXTITEM, TVGN_CARET, 0
        test    eax, eax
        jnz     .additfocused

        invoke  SendDlgItemMessageA, [.hwnd], ProjTreeCtrlID, TVM_GETNEXTITEM, TVGN_ROOT, 0
        test    eax, eax
        jnz     .additfocused

        stdcall OnAddCategory, eax, eax
        test    eax, eax
        jz      .erroradd
        jmp     .getcurrent

.additfocused:
        mov     [.lparam], eax

.addit:
        mov     esi, eax
        stdcall GetTVItemPtr, [.hwnd], esi
        stdcall AddArrayItems, eax, 1
        jc      .erroradd

        stdcall SetTVItemPtr, [.hwnd], esi, edx

        mov     ecx, [.wparam]
        mov     dword [eax], ecx
        mov     [ecx+TOpenFile.projectID], TRUE

        stdcall FileChanged, ecx, fcEditor

        invoke  SendDlgItemMessageA, [.hwnd], ProjTreeCtrlID, TVM_GETNEXTITEM, TVGN_CARET, 0
        cmp     eax, [.lparam]
        jne     .endadd

        stdcall ReloadListView, [.hwnd], [.lparam]

.endadd:
        xor     eax, eax
        clc
        return

.erroradd:
        xor     eax, eax
        inc     eax
        clc
        return

onmessage PMM_REMOVEFILE

        mov     [fProjModified], TRUE

        mov     esi, [.wparam]
        mov     [esi+TOpenFile.projectID], FALSE
        stdcall FileChanged, esi, fcEditor

        stdcall SearchForFile, [.hwnd], esi
        test    eax, eax
        jz      .mainfile
        js      .errdel

        mov     edi, eax
        mov     ebx, edx

        stdcall GetTVItemPtr, [.hwnd], edi
        stdcall DeleteArrayItems, eax, ebx, 1

        invoke  SendDlgItemMessageA, [.hwnd], ProjTreeCtrlID, TVM_GETNEXTITEM, TVGN_CARET, 0
        cmp     eax, edi
        jne     .enddel

        stdcall ReloadListView, [.hwnd], edi

.enddel:
        xor     eax, eax
        clc
        return

.mainfile:
        invoke  SendDlgItemMessageA, [.hwnd], ProjMainCtrlID, LVM_DELETEITEM, 0, 0
        jmp     .enddel

.errdel:
        xor     eax, eax
        inc     eax
        clc
        return

onmessage CEM_LOADFILE
locals
  .chunkname dd ?
endl
        mov     ebx, [ptrProjectFile]

        cmp     [.wparam], 0
        jne     .loadit

        cmp     [ebx+TOpenFile.fNeverSaved], FALSE
        je      .loadit

        mov     [fProjModified], FALSE
        xor     eax, eax
        clc
        return

.loadit:
        mov     [fProjModified], FALSE
        stdcall ChangeDir, [ebx+TOpenFile.hFileName]

; first set default options.
        stdcall EmptyVarList, poVarArray

; then load it...
        mov     eax, [.wparam]
        test    eax, eax
        jnz     @f
        mov     eax, [ebx+TOpenFile.hFileName]
@@:
        stdcall StrPtr, eax
        stdcall LoadBinaryFile, eax

        mov     [.ptr], eax
        mov     esi, eax

        cmp     ecx, 8
        jl      .invalidfile

; Project file header.
; Check the file singature
        lea     ecx, [ecx-8]

        lodsd
        cmp     eax, 'FNPr'   ; Fresh new project?
        jne     .invalidfile

        lodsd
        cmp     eax, $0a1a0a0d  ;) magic number. 'CR, LF, EOF, LF'
        je      .readchunks

.invalidfile:
        invoke  MessageBoxA, [hApplication], cErrorNotProject, NULL, MB_ICONERROR or MB_OK
        jmp     .endoffile

.readchunks:
        cmp     ecx, 8
        jl      .endoffile

        sub     ecx, 4
        lodsd
        mov     [.chunkname], eax

        call    .get_chunk_len
        jl      .invalidfile

        call    .check_chunk_hash
        jne     .invalidfile

; now esi points to the begin of the chunk data. edx contains the length of the chunk.
        push    ecx
        push    esi
        push    edx

        cmp     [.chunkname], 'PRTR'
        je      .read_project_tree

        cmp     [.chunkname], 'VARS'
        je      .read_vars

        cmp     [.chunkname], 'MAIN'
        je      .read_main_source_name

        cmp     [.chunkname], 'CMEM'
        je      .read_compiler_memory

        cmp     [.chunkname], 'CTNM'
        je      .read_output_filename

; unknown chunk. simply skip it.
.nextchunk:
        pop      edx
        pop      esi
        pop      ecx

        lea     esi, [esi+edx+4]  ; skip hash, it is already checked.
; adjust file length
        sub     ecx, edx
        sub     ecx, 4          ; 4 more bytes for the hash
        jmp     .readchunks


.endoffile:
        stdcall FreeMem, [.ptr]
        stdcall SetMainTitle

        stdcall CreateParametersMenu

        mov     ebx, [ptrProjectFile]
        stdcall AddToMRUList, [ptrMRUProj], [ebx+TOpenFile.hFileName]

        invoke  SendDlgItemMessageA, [.hwnd], ProjMainCtrlID, LVM_SETCOLUMNWIDTH, 0, LVSCW_AUTOSIZE

        xor     eax, eax
        clc
        return

  cErrorNotProject text 'Error: this file is not a project file in Fresh project format.',13, \
                          'New project created.'

;*****************************************************************************
;
; Chunk helper procedures.
;*****************************************************************************


;returns data length in EDX and datalen+4 in eax
.get_chunk_len:
        sub     ecx, 4
        lodsd   ; length of the chunk data
        mov     edx, eax
        add     eax, 4
        cmp     ecx, eax
        retn

; esi points to data.
; edx contains data len.
;
; returns computed hash in eax
.check_chunk_hash:
        push    esi edx ecx

        mov     eax, $811C9DC5                  ; 2166136261              ; FNV offset basis
        test    edx, edx
        jz      .endhash

.hashloop:
        movzx   ecx, byte [esi]
        inc     esi
        xor     eax, ecx
        imul    eax, $01000193                  ;   16777619              ; FNV prime
        dec     edx
        jnz     .hashloop

.endhash:
        mov     ecx, [esi]       ; hash from the chunk end
        cmp     eax, ecx
        pop     ecx edx esi
        retn

; esi points to the begin of the data.
; edi points just after the data.

.finalize_chunk:
        push    edx

        mov     edx, edi
        sub     edx, esi
        mov     [esi-4], edx      ; size of the chunk.
        stdcall .check_chunk_hash
        stosd                     ; save chunk hash at the end of the chunk.

        pop     edx
        retn


;--------------------------------------
; edi - pointer to the stream buffer
; eax - handle/pointer to the string
;--------------------------------------
.StrToStream:
        push    ecx
        stdcall StrPtr, eax
        xor     ecx, ecx

.copyname:
        mov     cl, [eax]
        inc     eax
        jecxz   .endofname
        mov     [edi], cl
        inc     edi
        jmp     .copyname

.endofname:
        pop     ecx
        retn






;***************************************************************************
;
; chunk 'PRTR' - contains one project category and its files.
;
;
.read_project_tree:
        lodsd                   ; icon number
        add     esi, 4          ; reserved dword for future use.

        stdcall GetCategoryTree, esi, TVI_ROOT, eax
        mov     ebx, eax        ; Current category tree item.

.filesloop:
        stdcall StrLen, esi
        lea     esi, [esi+eax+1]        ; goto to next file

        cmp     byte [esi], 0
        je      .nextchunk

        stdcall DetectFileType, esi
        stdcall CreateFStruct, esi, eax
        mov     edi, eax

        cmp     [.wparam], 0
        jne     .addtoproject    ; loaded from template.

        stdcall GetModified, edi
        test    eax, eax
        jz      .addtoproject

        stdcall ProjectInconsistent, edi
        cmp     eax, IDNO
        je      .adderror

.addtoproject:
        push    [fProjModified]
        invoke  SendMessageA, [.hwnd], PMM_ADDFILE, edi, ebx
        pop    [fProjModified]
;        mov     [fProjModified], FALSE

        test    eax, eax
        jz      .filesloop

.adderror:
        mov     [fProjModified], TRUE
        stdcall FreeFStruct, edi   ; error add file...
        jmp     .filesloop


;******************************************************************
.read_vars:
        stdcall StrLen, esi
        test    eax, eax
        jz      .nextchunk

        push    esi
        lea     esi, [esi+eax+1]
        stdcall StrDup ; from the stack.
        push    eax             ; key name

        stdcall StrLen, esi

        push    esi
        lea     esi, [esi+eax+1]
        stdcall StrDup ; from the stack.
        push    eax     ; key value.

        stdcall _AddKeyValueToArray, poVarArray ; two more arguments already in stack.
        jmp     .read_vars


;******************************************************************
; reads main source file name.

.read_main_source_name:

        stdcall StrLen, esi
        mov     ebx, eax

        test    ebx, ebx
        jz      .nextchunk

        stdcall CreateFStruct, esi, ftMainSource
        mov     edi, eax

        cmp     [.wparam], 0
        jne     .addmain        ; loaded from template so someone else have to bother for consistency

        stdcall GetModified, edi
        test    eax, eax
        jz      .addmain

        stdcall ProjectInconsistent, edi
        cmp     eax, IDNO
        jne     .addmain

        stdcall FreeFStruct, edi
        mov     [fProjModified], TRUE
        jmp     .nextchunk

.addmain:
        push   [fProjModified]
        invoke  SendMessageA, [.hwnd], PMM_ADDFILE, edi, NULL
        pop    [fProjModified]

        jmp     .nextchunk


;******************************************************************

.read_compiler_memory:
        lodsd
        mov     [poMemory], eax
        jmp     .nextchunk

;******************************************************************

.read_output_filename:
        xor     eax, eax
        lodsb
        mov     [poBinNameAuto], eax

        stdcall StrCopy, [poBinName], esi
        jmp     .nextchunk

;******************************************************************
onmessage _CEM_GETTEXT
        mov     eax, [ptrProjectFile]
        stdcall ChangeDir, [eax+TOpenFile.hFileName]

        mov     esi, [.lparam]  ; pointer to TInMemoryFile

        stdcall GetMem, 65536       ; Maximal size of the project file.
        jc      .errorgettext

        mov     [esi+TInMemoryFile.ptrBuffer], eax
        mov     edi, eax

; magic header.
        mov     eax, 'FNPr'
        stosd
        mov     eax, $0a1a0a0d  ;) magic number. 'CR, LF, EOF, LF'
        stosd

;_________________________
;
; write chunk MAIN
;_________________________

        invoke  SendMessageA, [.hwnd], PMM_GETMAINFILE, 0, 0
        test    eax, eax
        jz      .main_chunk_ok

        mov     ecx, eax
        mov     eax, 'MAIN'
        stosd

        add     edi, 4
        mov     esi, edi

        stdcall GetRelativePath, [ecx+TOpenFile.hFileName]
        push    eax
        call    .StrToStream
        stdcall StrDel ; from the stack
        xor     al,al
        stosb   ; end of string.

        call    .finalize_chunk
.main_chunk_ok:

;_________________________
;
; write chunks CTNM
;_________________________
        mov     eax, 'CTNM'     ; compiler target name.
        stosd

        add     edi, 4
        mov     esi, edi

        mov     al, byte [poBinNameAuto]
        stosb

        mov     eax, [poBinName]
        stdcall .StrToStream

        mov     al, 0
        stosb

        call    .finalize_chunk

;_________________________
;
; write chunks CMEM
;_________________________

        mov     eax, 'CMEM'
        stosd

        add     edi, 4
        mov     esi, edi

        mov     eax, [poMemory]
        stosd

        call    .finalize_chunk

;_________________________
;
; write chunks VARS
;_________________________

        mov     ebx, [poVarArray]
        test    ebx, ebx
        jz      .vars_ok

        mov     ecx, [ebx+TArray.count]
        lea     ebx, [ebx+TArray.array]
        jecxz   .vars_ok

        mov     eax, 'VARS'
        stosd

        add     edi, 4
        mov     esi, edi

.varsloop:
        mov     eax, [ebx+TEnvVarType.pName]
        call    .StrToStream
        mov     al, 0
        stosb
        mov     eax, [ebx+TEnvVarType.pValue]
        call    .StrToStream
        mov     al, 0
        stosb

        add     ebx, sizeof.TEnvVarType
        dec     ecx
        jnz     .varsloop

        stosb   ; final zero.

        call    .finalize_chunk

.vars_ok:

;_________________________
;
; write chunks PRTR
;_________________________

; Save categories...
        invoke  GetDlgItem, [.hwnd], ProjTreeCtrlID

        ;  proc CategoriesScan, hTree, tvRoot, lparam, ptrCallback
        stdcall CategoriesScan, eax, TVI_ROOT, eax, SaveOneCategory


; all is ok, so finish this work.
        mov     esi, [.lparam]
        sub     edi, [esi+TInMemoryFile.ptrBuffer]
        mov     [esi+TInMemoryFile.fileSize], edi
        xor     eax, eax
        mov     [esi+TInMemoryFile.fAscii], eax
        clc
        return

.errorgettext:
        mov     eax, 1
        stc
        return

;******************************************************************************************


onmessage  _CEM_CHANGENOTIFY
; change directory to the new name...
        mov     ebx, [ptrProjectFile]
        stdcall ChangeDir, [ebx+TOpenFile.hFileName]
        stdcall SetMainTitle
        stdcall AddToMRUList, [ptrMRUProj], [ebx+TOpenFile.hFileName]
        clc
        return


onmessage _CEM_FOCUSFILE
        invoke  SetFocus, [.hwnd]
        clc
        return


onmessage _CEM_CANACTION
        xor     eax, eax
        clc
        return

onmessage _CEM_EDITACTION
        clc
        return


onmessage _CEM_GETFILE
        mov     eax, [ptrProjectFile]
        clc
        return


onmessage _CEM_GETMODIFIED
        mov     eax, [fProjModified]
        clc
        return


onmessage _CEM_SETMODIFIED
        push    [.wparam]
        pop     [fProjModified]
        clc
        return


onmessage WM_ACTIVATE
        stdcall FocusFile, [ptrProjectFile]
        stc
        return

onmessage WM_DROPFILES
        lea     esi, [.str]
        invoke  DragQueryFile, [.wparam], -1, esi, 1024
        mov     ebx, eax
        xor     edi, edi

.opendroploop:
        invoke  DragQueryFile, [.wparam], edi, esi, 1024
        stdcall OpenFileByName, esi, NULL, FALSE
        test    eax, eax
        jz      .next
        invoke  SendMessageA, [.hwnd], PMM_ADDFILE, eax, -1
.next:
        inc     edi
        dec     ebx
        jnz     .opendroploop

        invoke  DragFinish, [.wparam]

        xor     eax, eax
        clc
        return

onmessage WM_COMMAND
        mov     eax, [.wparam]

        cmp     eax, ProjSettingsID
        je      .projsettings

        cmp     eax, IDOK
        je      .enter_key

        cmp     eax, FirstID
        jb      .from_menu

.tomain:
        invoke  PostMessageA, [hMainWindow], WM_COMMAND, [.wparam], [.lparam]

        clc
        xor     eax, eax
        return


.projsettings:
locals
  .btnrect RECT
endl
        invoke  GetDlgItem, [.hwnd], ProjSettingsID
        lea     ebx, [.btnrect]
        invoke  GetWindowRect, eax, ebx
        invoke  TrackPopupMenu, [hSettingsPopup], TPM_LEFTALIGN or TPM_LEFTBUTTON, \
                [.btnrect.left], [.btnrect.top], 0, [.hwnd], 0

        clc
        return


.enter_key:
        invoke  GetFocus
        invoke  GetDlgCtrlID, eax

        cmp     eax, ProjListCtrlID
        je      .listviewenter
        cmp     eax, ProjMainCtrlID
        jne     .ondefault

.listviewenter:
        mov     [.wparam], eax
        jmp     .itemactivate


.from_menu:
; this is actually fast settings menu.
        movzx   eax, word [.wparam]

        shr     eax, 8
        sub     eax, VarMenuBaseID
        mov     esi, [poVarArray]
        cmp     eax, [esi+TArray.count]
        jae     .endmenu

        lea     esi, [esi+eax*sizeof.TEnvVarType + TArray.array]
; so process the value.
        movzx   ecx, word [.wparam]
        and     ecx, $ff

        stdcall StrPtr, [esi+TEnvVarType.pValue]
        mov     edi, eax
        mov     edx, eax

.loop1:
        cmp     byte [edi], 0
        je      .endmenu

        cmp     byte [edi], '|'
        lea     edi, [edi+1]
        jne     .loop1
        loop    .loop1

        push    edi     ; first separator + 1

.loop2:
        mov     al, [edi]
        lea     edi, [edi+1]
        test    al, al
        jz      .endfound
        cmp     al, '|'
        jne     .loop2

.endfound:
        movzx   ecx, al
        mov     byte [edi-1], 0

        pop     esi
        stdcall StrDup, esi
        mov     ebx, eax

        mov     [edi-1], cl       ; restore the separator.
        sub     edi, 2
        sub     esi, 2

.shift:
        cmp     esi, edx
        jb      .endshift

        mov     al, [esi]
        lea     esi, [esi-1]
        mov     [edi], al
        lea     edi, [edi-1]
        jmp     .shift

.endshift:
        mov     byte [edi], '|'
        mov     edi, edx

        stdcall StrPtr, ebx
        mov     esi, eax

.copyloop:
        mov     al, [esi]
        lea     esi, [esi+1]

        test    al, al
        jz      .endcopy

        mov     [edi], al
        lea     edi, [edi+1]
        jmp     .copyloop

.endcopy:
        stdcall StrDel, ebx
        stdcall CreateParametersMenu
        invoke  SendMessageA, [.hwnd], _CEM_SETMODIFIED, TRUE, 0

.endmenu:
        clc
        return



;**********************************************************************************************

onmessage PMM_GETMAINFILE

        mov     [.lvi.mask], LVIF_PARAM
        mov     [.lvi.iItem], 0
        lea     eax, [.lvi]
        invoke  SendDlgItemMessageA, [.hwnd], ProjMainCtrlID, LVM_GETITEM, 0, eax
        test    eax, eax
        jz      .endgetmain
        mov     eax, [.lvi.lParam]
.endgetmain:
        clc
        return

;**********************************************************************************************

onmessage WM_NOTIFY
        mov     ecx, [.wparam]
        mov     esi, [.lparam]
        cmp     ecx, ProjTreeCtrlID
        je      .treenotify
        cmp     ecx, ProjMainCtrlID
        je      .listnotify
        cmp     ecx, ProjListCtrlID
        je      .listnotify

        stc
        return

;-----------ListView notification messages-----------------

.listnotify:
        cmp     [esi+NMHDR.code], LVN_ITEMACTIVATE
        je      .itemactivate
        cmp     [esi+NMHDR.code], LVN_KEYDOWN
        je      .listkey
        cmp     [esi+NMHDR.code], NM_RCLICK
        je      .lvrightclick

        stc
        return


.lvrightclick:
        push    [hListPopup]

        lea     eax, [.pnt]
        invoke  GetCursorPos, eax
        pop     eax

        invoke  TrackPopupMenu, eax, TPM_LEFTALIGN or TPM_RIGHTBUTTON, [.pnt.x], [.pnt.y], 0, [.hwnd], NULL
        jmp     .qfalse

.itemactivate:
        xor     ebx, ebx
        dec     ebx

.openloop:
        invoke  SendDlgItemMessageA, [.hwnd], [.wparam], LVM_GETNEXTITEM, ebx, LVNI_SELECTED
        cmp     eax, -1
        je      .qfalse

        mov     ebx, eax

        mov     [.lvi.mask], LVIF_PARAM
        mov     [.lvi.iItem], eax
        lea     eax, [.lvi]
        invoke  SendDlgItemMessageA, [.hwnd], [.wparam], LVM_GETITEM, 0, eax
        stdcall LoadFile, [.lvi.lParam], NULL
        jmp     .openloop


;---------- Del key is pressed in the list view -------------------
.listkey:
        cmp     [esi+LVKEYDOWN.wVKey], VK_DELETE
        jne     .qfalse

.delitems:
        xor     ebx, ebx
        push    ebx             ; marker for the end of the list
        dec     ebx
.delloop:
        invoke  SendDlgItemMessageA, [.hwnd], [.wparam], LVM_GETNEXTITEM, ebx, LVNI_SELECTED
        cmp     eax, -1
        je      .delit

        mov     ebx, eax

        mov     [.lvi.mask], LVIF_PARAM
        mov     [.lvi.iItem], eax
        lea     eax, [.lvi]
        invoke  SendDlgItemMessageA, [.hwnd], [.wparam], LVM_GETITEM, 0, eax
        push    [.lvi.lParam]
        jmp     .delloop

.delit:
        cmp     dword [esp], 0
        je      .delitloop

        invoke  MessageBoxA, [hApplication], cDelFileQuery, cDelFileTitle, MB_ICONQUESTION or MB_YESNO
        cmp     eax, IDYES
        je      .delitloop

.clearstack:
        pop     ecx
        jecxz   .endfiledel
        jmp     .clearstack

.delitloop:
        pop     ecx
        jecxz   .endfiledel
        invoke  SendMessageA, [.hwnd], PMM_REMOVEFILE, ecx, 0
        jmp     .delitloop

.endfiledel:
        xor     eax, eax
        clc
        return

  cDelFileQuery text 'Do you really want to remove selected file(s) from the project?'
  cDelFileTitle text 'Remove file from project.'

;--- Category tree notification messages ---------
.treenotify:
        cmp     [esi+NMHDR.code], TVN_SELCHANGED
        je      .selchanged

        cmp     [esi+NMHDR.code], TVN_KEYDOWN
        je      .keydown

        cmp     [esi+NMHDR.code], NM_RCLICK
        je      .tvrightclick

        cmp     [esi+NMHDR.code], TVN_ENDLABELEDIT
        je      .endlabeledit

        cmp     [esi+NMHDR.code], TVN_DELETEITEM
        jne     .qfalse

.deleteitem:
        mov     edi, [esi+NMTREEVIEW.itemOld.lParam]
        test    edi, edi
        jz      .qfalse

        test    [esi+NMTREEVIEW.itemOld.state], TVIS_FOCUSED
        jz      .freelist

        invoke  SendDlgItemMessageA, [.hwnd], ProjListCtrlID, LVM_DELETEALLITEMS, 0, 0

.freelist:
        stdcall ListFree, edi, FileListFreeCallback
        jmp     .qfalse

;----------- end of label editing ------------------------
.endlabeledit:
        mov     esi, [.lparam]
        cmp     [esi+TVDISPINFO.item.pszText], 0
        je      .endedit

        mov     [fProjModified], TRUE
        mov     [esi+TVDISPINFO.item.mask], TVIF_TEXT
        lea     esi, [esi+TVDISPINFO.item]
        invoke  SendDlgItemMessageA, [.hwnd], ProjTreeCtrlID, TVM_SETITEM, 0, esi

.endedit:
       clc
       return

;---------- right click on treeview control --------------
.tvrightclick:
        lea     eax, [.pnt]
        invoke  GetCursorPos, eax
        invoke  TrackPopupMenu, [hTreePopup], TPM_LEFTALIGN or TPM_RIGHTBUTTON, [.pnt.x], [.pnt.y], 0, [.hwnd], NULL
        jmp     .qfalse


;---------- selection changed notification ---------------
.selchanged:
; load the new selected category in the listview...
        stdcall ReloadListView, [.hwnd], [esi+NMTREEVIEW.itemNew.hItem]
        xor     eax, eax
        clc
        return

;------------- key down notification -----------------
.keydown:
        cmp     [esi+TVKEYDOWN.wVKey], VK_DELETE
        jne     .qfalse

; delete key on category...
        invoke  MessageBoxA, [hApplication], cDelProjQuery, cDelProjTitle, MB_ICONQUESTION or MB_YESNO
        cmp     eax, IDYES
        jne     .qfalse

        invoke  SendDlgItemMessageA, [.hwnd], ProjTreeCtrlID, TVM_GETNEXTITEM, TVGN_CARET, 0
        invoke  SendDlgItemMessageA, [.hwnd], ProjTreeCtrlID, TVM_DELETEITEM, 0, eax
        jmp     .qfalse

cDelProjQuery text 'Do you really want to remove this category and all category files from the project?'
cDelProjTitle text 'Remove category.'

onmessage WM_CLOSE
        stdcall CloseFile, [ptrProjectFile]
        cmp     eax, cfrDestroyed
        jne     .qfalse
        mov     [ptrProjectFile], 0
        jmp     .qfalse

onmessage WM_DESTROY
        mov     [ptrProjectFile], 0

        invoke  SendMessageA, [.hwnd], PMM_GETMAINFILE, 0, 0
        test    eax, eax
        jz      .mainfree

        stdcall FileListFreeCallback, eax

.mainfree:
        stdcall DeleteAllTVItems, [hProjManager] ; clear all items in the project and frees the memory.
        mov     [hProjManager], 0

        stdcall StrDel, [poBinName]
        mov     [poBinName], 0
        mov     [poBinNameAuto], TRUE

        cmp     [poVarArray], 0
        je      .qfalse

        stdcall FreeMem, [poVarArray]
        mov     [poVarArray], 0

        jmp     .qfalse


cMRUProjSection text 'MRU projects'

onmessage AM_INITWINDOW

        xor     ebx, ebx

        invoke  CreateWindowExA, 0, cButtonClassName, ebx,                   \
                                WS_CHILD or WS_VISIBLE or BS_OWNERDRAW or   \
                                BS_VCENTER or WS_TABSTOP,                   \
                                ebx, ebx, ebx, 16,                          \
                                [.hwnd], ProjSettingsID, [hInstance], ebx

        push    0 BST_CHECKED BM_SETCHECK eax
        stdcall SetAlign, eax, waTop
        invoke  SendMessageA ; from the stack

        invoke  CreateWindowExA, WS_EX_CLIENTEDGE, cListviewClassName, ebx,      \
                                WS_CHILD or WS_VISIBLE or LVS_REPORT or         \
                                LVS_SHAREIMAGELISTS or LVS_SORTASCENDING or     \
                                LVS_NOSCROLL or LVS_NOCOLUMNHEADER or           \
                                LVS_NOSORTHEADER or WS_TABSTOP,                 \
                                ebx, ebx, ebx, 20,                              \
                                [.hwnd], ProjMainCtrlID, [hInstance], ebx

        mov     edi, eax
        stdcall SetAlign, eax, waTop

        mov     [.col.mask], LVCF_FMT or LVCF_WIDTH
        mov     [.col.fmt], LVCFMT_LEFT
        mov     [.col.cx], 1024
        lea     eax, [.col]
        invoke  SendMessageA, edi, LVM_INSERTCOLUMN, ebx, eax
        invoke  SendMessageA, edi, LVM_SETIMAGELIST, LVSIL_SMALL, [hTabIml]

        invoke  CreateWindowExA, ebx, cStaticClassName, ebx,              \
                                WS_CHILD or WS_VISIBLE or SS_LEFT,       \
                                ebx, ebx, ebx, 4,                        \
                                [.hwnd], ebx, [hInstance], ebx
        stdcall SetAlign, eax, waTop

        invoke  CreateWindowExA, WS_EX_CLIENTEDGE,                                                       \
                                cTreeViewClassName, ebx,                                                \
                                WS_CHILD or WS_VISIBLE or TVS_DISABLEDRAGDROP or TVS_HASBUTTONS or     \
                                TVS_SHOWSELALWAYS or TVS_HASLINES or TVS_LINESATROOT or TVS_EDITLABELS or WS_TABSTOP, \
                                ebx, ebx, ebx, 100,                                                     \
                                [.hwnd], ProjTreeCtrlID, [hInstance], ebx
        mov     edi, eax
        stdcall SetAlign, eax, waTop
        invoke  SendMessageA, edi, TVM_SETIMAGELIST, TVSIL_NORMAL, [imlCategories]

        stdcall CreateSplitter, edi, TRUE, resVSplit, resHSplit
        invoke  SendMessageA, eax, SPM_SETMAXMIN, 16, 1000


        invoke  CreateWindowExA, WS_EX_CLIENTEDGE, cListviewClassName, ebx,              \
                                WS_CHILD or WS_VISIBLE or LVS_REPORT or                 \
                                LVS_SHAREIMAGELISTS or LVS_SORTASCENDING or            \
                                LVS_NOCOLUMNHEADER or                                  \
                                LVS_SHOWSELALWAYS or WS_TABSTOP or LVS_NOSORTHEADER,  \
                                ebx, ebx, ebx, ebx,                                     \
                                [.hwnd], ProjListCtrlID, [hInstance], ebx

        mov     edi, eax
        stdcall SetAlign, eax, waClient

        invoke  SendMessageA, edi, LVM_SETIMAGELIST, LVSIL_SMALL, [hTabIml]
        invoke  SendMessageA, edi, LVM_SETEXTENDEDLISTVIEWSTYLE, 0, \
                             LVS_EX_LABELTIP


        mov     [.col.mask], LVCF_FMT or LVCF_SUBITEM
        mov     [.col.fmt], LVCFMT_LEFT
        mov     [.col.iSubItem], 0
        lea     eax, [.col]
        invoke  SendMessageA, edi, LVM_INSERTCOLUMN, ebx, eax
        invoke  DragAcceptFiles, [.hwnd], TRUE

.qfalse:
        xor     eax,eax
        clc
        return


onmessage WM_DRAWITEM
        cmp     [.wparam], ProjSettingsID
        jne     .ondefault

        mov     esi, [.lparam]
        mov     edi, [esi+DRAWITEMSTRUCT.hDC]
        lea     ebx, [esi+DRAWITEMSTRUCT.rcItem]

        invoke  SetBkMode, edi, TRANSPARENT

        invoke  GetSysColor, COLOR_BTNFACE
        invoke  CreateSolidBrush, eax
        push    eax
        invoke  FillRect, edi, ebx, eax
        invoke  DeleteObject ; from the stack.

        invoke  DrawEdge, edi, ebx, EDGE_ETCHED,  BF_RECT; or BF_FLAT

        stdcall StrPtr, [hSettingsString]
        test    eax, eax
        je      .setshort
        cmp     eax, -1
        jne     .drawit

.setshort:
        mov     eax, cSettingsCaption

.drawit:
        add     [ebx+RECT.left],4
        invoke  DrawTextA, edi, eax, -1, ebx, DT_LEFT or DT_VCENTER or DT_SINGLELINE

        clc
        mov     eax, 1
        return

endwp




; it should return CF=1 if the result is still not found.
; arguments are [ptrList], [hTVItem] and [.lparam]
proc SaveOneCategory, .ptrList, .hItem, .hTree
.tvi TVITEM
.count dd ?
begin
        push    ebx esi

        mov     eax, 'PRTR'
        stosd

        add     edi, 4
        mov     esi, edi        ; pointer to the data.

        stdcall GetFullItemName, [.hTree], [.hItem]
        mov     ebx, eax

        mov     [.tvi.mask], TVIF_IMAGE or TVIF_PARAM
        mov     eax, [.hItem]
        mov     [.tvi.hItem], eax
        lea     eax, [.tvi]
        invoke  SendMessageA, [.hTree], TVM_GETITEM, 0, eax
        test    eax, eax
        jz      .finish

        mov     eax, [.tvi.iImage]         ; image index.
        stosd
        xor     eax, eax                   ; reserved.
        stosd

        mov     eax, ebx
        call    ProjManProc.StrToStream
        xor     al,al
        stosb

        stdcall StrDel, ebx     ; category name.

        mov     ebx, [.tvi.lParam]
        mov     edx, [ebx+TArray.count]
        mov     [.count], edx
        lea     ebx, [ebx+TArray.array]

.looplist:
        dec     [.count]
        js      .finish

        mov     eax, [ebx]      ; ptr to TOpenFile
        lea     ebx, [ebx+4]

        stdcall GetRelativePath, [eax+TOpenFile.hFileName]
        push    eax
        call    ProjManProc.StrToStream
        stdcall StrDel ; from the stack
        xor     al,al
        stosb
        jmp     .looplist

.finish:
        xor     eax, eax
        stosb

        call    ProjManProc.finalize_chunk

        stc
        pop     esi ebx
        return
endp



proc FileListFreeCallback, .ptrOpenFile
begin
        push    esi
        mov     esi, [.ptrOpenFile]

        mov     [esi+TOpenFile.projectID], FALSE
        stdcall FileChanged, esi, fcEditor
        pop     esi
        return
endp


;***********************************************************
; Set the filename of the project and updates the title of
; the main window.
;***********************************************************
proc SetMainTitle
begin
        push    esi

        xor     eax, eax
        mov     esi, [ptrProjectFile]
        test    esi, esi
        jz      @f

        stdcall StrPtr, [esi+TOpenFile.hFileName]

@@:
        invoke  SendMessageA, [hMainWindow], WM_SETTEXT, 0, eax
        pop     esi
        return
endp



;***********************************************************
; List view utils.
;***********************************************************
proc ReloadListView, .hProj, .hCategory
begin
        push    edi ebx

        invoke  SendDlgItemMessageA, [.hProj], ProjListCtrlID, LVM_DELETEALLITEMS, 0, 0
        stdcall GetTVItemPtr, [.hProj], [.hCategory]
        mov     edi, eax

        mov     eax, [edi+TArray.count]
        mov     ebx, eax
        invoke  SendDlgItemMessageA, [.hProj], ProjListCtrlID, LVM_SETITEMCOUNT, ebx, 0

        lea     edi, [edi+TArray.array]

.addnewlist:
        dec     ebx
        js      .finish

        stdcall AddToListView, [.hProj], ProjListCtrlID, [edi]
        lea     edi, [edi+4]
        jmp     .addnewlist

.finish:
        invoke  GetDlgItem, [.hProj], ProjListCtrlID
        invoke  PostMessageA, eax, LVM_SETCOLUMNWIDTH, 0, LVSCW_AUTOSIZE
        pop     ebx edi
        return
endp


proc AddToListView, .hProject, .ListID, .ptrOpenFile
.lvi LVITEM
begin
        push    esi
        mov     esi, [.ptrOpenFile]

        stdcall GetScreenProperties, esi
        test    eax, eax
        jz      .finish

        push    eax     ; for StrDel
        stdcall StrPtr, eax

        mov     [.lvi.mask], LVIF_IMAGE or LVIF_PARAM or LVIF_TEXT
        mov     [.lvi.iItem], 0
        mov     [.lvi.iSubItem], 0
        mov     [.lvi.iImage], edx
        mov     [.lvi.lParam], esi
        mov     [.lvi.pszText], eax

        lea     eax, [.lvi]
        invoke  SendDlgItemMessageA, [.hProject], [.ListID], LVM_INSERTITEM, 0, eax
        stdcall StrDel ; from the stack.
.finish:
        pop     esi
        return
endp


;------------------------------------------------------------;
; Searches the project for the file and returns:             ;
; eax - handle of the category that contain this file.       ;
;   eax = 0 if the file is main file.                        ;
;   eax = -1 if the file was not found.                      ;
; edx - index in TList containing the file pointer.          ;
;------------------------------------------------------------;
proc SearchForFile, .hProject, .ptrOpenFile
begin
        invoke  SendMessageA, [.hProject], PMM_GETMAINFILE, 0, 0
        test    eax, eax
        jz      .nomain

        cmp     eax, [.ptrOpenFile]
        je      .main

.nomain:
        invoke  GetDlgItem, [.hProject], ProjTreeCtrlID
        mov     ebx, eax
        stdcall CategoriesScan, ebx, TVI_ROOT, [.ptrOpenFile], _SearchCategory
        return

.main:
        mov     edx, eax
        xor     eax, eax
        return
endp


proc _SearchCategory, .ptrList, .hTVItem, .ptrOpenFile
begin
        stdcall ListIndexOf, [.ptrList], [.ptrOpenFile]
        jc      .finish

        mov     edx, [.hTVItem]
        xchg    eax, edx
        clc
.finish:
        return
endp


;---------------------------------------------------------------
; Recursively loops through all categories and calls callback
; procedure for every category.
;---------------------------------------------------------------
proc CategoriesScan, .hTree, .tvRoot, .lparam, .ptrCallback
.tvi TVITEM
begin
        push    esi ebx
        invoke  SendMessageA, [.hTree], TVM_GETNEXTITEM, TVGN_CHILD, [.tvRoot]

.loop:
        mov     ebx, eax
        test    ebx, ebx
        jz      .notfound

        mov     [.tvi.mask], TVIF_PARAM
        mov     [.tvi.hItem], ebx
        lea     eax, [.tvi]
        invoke  SendMessageA, [.hTree], TVM_GETITEM, 0, eax

        mov     esi, [.tvi.lParam]
        test    esi, esi
        jz      .listdone

; it should return CF=1 if the result is still not found.
; arguments are [ptrList] and [hTVItem] and [.lparam]
        stdcall [.ptrCallback], esi, ebx, [.lparam]
        jnc     .found

.listdone:
        stdcall CategoriesScan, [.hTree], ebx, [.lparam], [.ptrCallback]
        jnc     .found

        invoke  SendMessageA, [.hTree], TVM_GETNEXTITEM, TVGN_NEXT, ebx
        jmp     .loop

.notfound:
        stc
        pop     ebx esi
        return

.found:
        pop     ebx esi
        return
endp


;---------------------------------------------
; Returns handle to string contains full path
; name for given tree-view item.
; returns also, the image index of the item in ecx
;---------------------------------------------
proc GetFullItemName, .hTree, .hItem
.tvi TVITEM
.str rb 256
.img dd ?
begin
        push    esi ebx

        stdcall StrNew
        mov     esi, eax
        mov     eax, [.hItem]

.buildloop:
        mov     ebx, eax
        test    ebx, ebx
        jz      .rootoftree

        mov     [.tvi.mask], TVIF_TEXT or TVIF_IMAGE
        mov     [.tvi.hItem], ebx
        lea     eax, [.str]
        mov     [.tvi.pszText], eax
        mov     [.tvi.cchTextMax], 256
        lea     eax, [.tvi]
        invoke  SendMessageA, [.hTree], TVM_GETITEM, 0, eax

        cmp     ebx, [.hItem]
        jne     .dostring

        push    [.tvi.iImage]
        pop     [.img]

.dostring:
         stdcall StrCharInsert, esi, '\', 0
        stdcall StrInsert, esi, [.tvi.pszText], 0

        invoke  SendMessageA, [.hTree], TVM_GETNEXTITEM, TVGN_PARENT, ebx
        jmp     .buildloop

.rootoftree:
        mov     eax, esi
        mov     ecx, [.img]
        pop     ebx esi
        return
endp


;***********************************************************
; TreeView utils. This is small library for easy work with
; TreeView control.
;***********************************************************

;**************************************************************
; Returns lParam of the TreeView item.
;**************************************************************
proc GetTVItemPtr, .hProject, .hItem
.item   TVITEM
begin
        push    ecx edx

        push    [.hItem]
        pop     [.item.hItem]
        mov     [.item.mask], TVIF_PARAM
        lea     eax, [.item]
        invoke  SendDlgItemMessageA, [.hProject], ProjTreeCtrlID, TVM_GETITEM, 0, eax
        test    eax, eax
        jz      .error

        mov     eax, [.item.lParam]
        clc
        pop     edx ecx
        return

.error:
        stc
        pop     edx ecx
        return
endp



;**************************************************************
; Sets lParam of the TreeView item.
;**************************************************************
proc SetTVItemPtr, .hProject, .hItem, .lparam
.item   TVITEM
begin
        push    eax ecx edx
        push    [.hItem]
        pop     [.item.hItem]
        mov     [.item.mask], TVIF_PARAM
        push    [.lparam]
        pop     [.item.lParam]
        lea     eax, [.item]
        invoke  SendDlgItemMessageA, [.hProject], ProjTreeCtrlID, TVM_SETITEM, 0, eax
        pop     edx ecx eax
        return
endp





;***********************************************
; Handler of action actMoveCategory
;***********************************************
proc OnMoveCategory, .wparam, .lparam
begin
        xor     ebx, ebx
        cmp     [.wparam], actMoveCategoryUp
        je      @f
        inc     ebx
@@:
        invoke  GetDlgItem, [hProjManager], ProjTreeCtrlID
        mov     esi, eax

        invoke  SendMessageA, esi, TVM_GETNEXTITEM, TVGN_CARET, 0
        stdcall MoveTVItem, esi, eax, ebx
        mov     [fProjModified], TRUE
        return
endp


;****************************************************************
; Moves one hItem up or down in the treeview, depending of the
; fMoveDown:
;    FALSE - move up
;    TRUE  - move down
;****************************************************************
proc MoveTVItem, .hTreeView, .hItem, .fMoveDown
.ptr dd ?
.obj dd ?
.tvi TVITEM
.lparam dd ?
.tvs TVSORTCB
begin
        push    esi edi ebx

        invoke  SendMessageA, [.hTreeView], TVM_GETCOUNT, 0, 0
        shl     eax, 2
        stdcall GetMem, eax
        mov     [.ptr], eax
        mov     edi, eax

        mov     eax, [.hItem]
        mov     [.tvi.hItem], eax
        mov     [.tvi.mask], TVIF_PARAM
        lea     eax, [.tvi]
        invoke  SendMessageA, [.hTreeView], TVM_GETITEM, 0, eax
        mov     eax, [.tvi.lParam]
        mov     [.lparam], eax

        invoke  SendMessageA, [.hTreeView], TVM_GETNEXTITEM, TVGN_PARENT, [.hItem]
        mov     ebx, eax

        invoke  SendMessageA, [.hTreeView], TVM_GETNEXTITEM, TVGN_CHILD, ebx
.getloop:
        test    eax, eax
        jz      .endofchildren

        mov     [.tvi.hItem], eax
        mov     [.tvi.mask], TVIF_PARAM or TVIF_HANDLE
        lea     eax, [.tvi]
        invoke  SendMessageA, [.hTreeView], TVM_GETITEM, 0, eax
        mov     eax, [.tvi.lParam]
        cmp     eax, [.lparam]
        jne     .storeit
        mov     [.obj], edi
.storeit:
        stosd

        invoke  SendMessageA, [.hTreeView], TVM_GETNEXTITEM, TVGN_NEXT, [.tvi.hItem]
        jmp     .getloop

.endofchildren:
        stosd
        cmp     [.fMoveDown], FALSE
        je      .moveup

.movedown:
        mov     edi, [.obj]
        mov     eax, [edi+4]
        test    eax, eax
        jz      .cantmove       ; it is the last element.

        mov     ecx, [edi]
        mov     [edi], eax
        mov     [edi+4], ecx
        jmp     .sortit

.moveup:
        mov     edi, [.obj]
        cmp     edi, [.ptr]
        je      .cantmove               ; it is the first element

        mov     eax, [edi-4]
        mov     ecx, [edi]
        mov     [edi], eax
        mov     [edi-4], ecx

.sortit:
        mov     [.tvs.hParent], ebx
        mov     [.tvs.lpfnCompare], CompareTwoItems
        mov     eax, [.ptr]
        mov     [.tvs.lParam], eax
        lea     eax, [.tvs]
        invoke  SendMessageA, [.hTreeView], TVM_SORTCHILDRENCB, 0, eax

.cantmove:
        stdcall FreeMem, [.ptr]
        pop     ebx edi esi
        return
endp


proc CompareTwoItems, .lParam1, .lParam2, .ptrList
begin
        push    esi

        mov     esi, [.ptrList]
        mov     ecx, [.lParam1]
        mov     edx, [.lParam2]
.loop:
        lodsd
        test    eax, eax
        jz      .notfound

        cmp     eax, ecx
        je      .first

        cmp     eax, edx
        jne     .loop

.second:
        xor     eax, eax
        inc     eax
        pop     esi
        return

.first:
        xor     eax, eax
        dec     eax
        pop     esi
        return

.notfound:
        xor     eax, eax
        pop     esi
        return
endp


;****************************************************************
; Returns in ebx the handle of parent treeview item.
;****************************************************************
proc GetItemParent, .hItem
begin
        push    eax ecx edx

        invoke  SendDlgItemMessageA, [hProjManager], ProjTreeCtrlID, TVM_GETNEXTITEM, TVGN_PARENT, [.hItem]
        test    eax,eax
        jz      .error
        clc
        jmp     .finish

.error:
        stc
.finish:
        mov     ebx, eax
        pop     edx ecx eax
        return
endp


proc DeleteAllTVItems, .hProj
begin
.loop:
        invoke  SendDlgItemMessageA, [.hProj], ProjTreeCtrlID, TVM_GETNEXTITEM, TVGN_ROOT, 0
        test    eax, eax
        jz      .finish

        stdcall DeleteOneTVItem, [.hProj], eax
        jmp     .loop
.finish:
        return
endp


proc DeleteOneTVItem, .hProj, .hItem
begin
.clearloop:
        invoke  SendDlgItemMessageA, [.hProj], ProjTreeCtrlID, TVM_GETNEXTITEM, TVGN_CHILD, [.hItem]
        test    eax, eax
        jz      .childrendeleted

        stdcall DeleteOneTVItem, [.hProj], eax
        jmp     .clearloop

.childrendeleted:
        invoke  SendDlgItemMessageA, [.hProj], ProjTreeCtrlID, TVM_DELETEITEM, 0, [.hItem]
        return
endp


;-------------------------------------------------------------------------
; Returns the handle of treeview item that follow the tree of direcories
; of [hFileName]. If there is no some (or any) of the directories of the
; path, it will be created.
; The function checks for "/" and "\" as items separator.
; Arguments:
; hFileName - handle or pointer to the path.
; hRoot     - is root TVItem for category. For examples if category
;             have to be created in the root of the TreeView, hParent
;             should be TVI_ROOT
; Returns:
;   eax - Handle to treeview item that should be parent of the file.
;   edx - pointer to the name of the file without path.
;-------------------------------------------------------------------------
proc GetCategoryTree, .hFileName, .hRoot, .iImage
begin
        push    esi edi ebx



        stdcall StrPtr, [.hFileName]
        mov     esi, eax

        stdcall StrLen, [.hFileName]
        mov     ecx, eax

; find the filename:
.findloop:
        cmp     byte [esi+ecx], '/'
        je      .foundname
        cmp     byte [esi+ecx], '\'
        je      .foundname
        dec     ecx
        jns     .findloop

; Returns the [hRoot] if there is no directories names.
        mov     eax, [.hRoot]
        mov     edx, esi
        pop     ebx edi esi
        return

.foundname:
        lea     edi, [esi+ecx+1]
        pushd   [edi]
        and     dword [edi], 0
        stdcall _DoCategoryTree, esi, [.hRoot], [.iImage]    ; track/create path items.
        popd    [edi]
        mov     edx, edi
        pop     ebx edi esi
        return
endp


proc _DoCategoryTree, .ptrName, .hTVRoot, .iImage
begin
        push    esi edi ebx
        mov     esi, [.ptrName]
        mov     ebx, [.hTVRoot]

.outloop:
        cmp     byte [esi], 0   ; is it the end?
        je      .endloop
        mov     edi, esi
.loop:
        lodsb
        cmp     al,'\'
        je      .foundslash
        cmp     al,'/'
        jne     .loop

.foundslash:
        mov     byte [esi-1], 0
        stdcall _DoGetOneItem, edi, ebx, [.iImage]
        mov     ebx, eax
        mov     byte [esi-1], '/'
        jmp     .outloop

.endloop:
        mov     eax, ebx
        pop     ebx edi esi
        return
endp


proc _DoGetOneItem, .ptrItem, .hParent, .iImage
.tvi TVITEM
.tvins TVINSERTSTRUCT
.str rb 256
begin
        push    esi edi ebx

        invoke  SendDlgItemMessageA, [hProjManager], ProjTreeCtrlID, TVM_GETNEXTITEM, TVGN_CHILD, [.hParent]
        test    eax, eax
        jz      .notfound
.itemscan:
        mov     ebx, eax
        mov     [.tvi.hItem], eax
        lea     esi, [.str]
        mov     [.tvi.pszText], esi
        mov     [.tvi.cchTextMax], 256
        mov     [.tvi.mask], TVIF_HANDLE or TVIF_TEXT
        lea     ecx, [.tvi]
        invoke  SendDlgItemMessageA, [hProjManager], ProjTreeCtrlID, TVM_GETITEM, 0, ecx

        stdcall StrCompNoCase, esi, [.ptrItem]
        jc      .found

        invoke  SendDlgItemMessageA, [hProjManager], ProjTreeCtrlID, TVM_GETNEXTITEM, TVGN_NEXT, ebx
        test    eax, eax
        jnz     .itemscan

.notfound:
        mov     eax, [.hParent]
        mov     [.tvins.hParent], eax
        mov     [.tvins.hInsertAfter], TVI_LAST
        mov     [.tvins.item.mask], TVIF_TEXT or TVIF_IMAGE or TVIF_SELECTEDIMAGE or TVIF_PARAM
        mov     eax, [.iImage]
        mov     [.tvins.item.iImage], eax
        mov     [.tvins.item.iSelectedImage], eax

        mov     eax, [.ptrItem]
        mov     [.tvins.item.pszText], eax

        stdcall CreateArray, 4
        mov     [.tvins.item.lParam], eax

        lea     eax, [.tvins]
        invoke  SendDlgItemMessageA, [hProjManager], ProjTreeCtrlID, TVM_INSERTITEM, 0, eax
        pop     ebx edi esi
        return

.found:
        mov     eax, ebx
        pop     ebx edi esi
        return
endp



;--------------------------------------------------------------------
; Recursively scans given source file [hFileName] for "include"
; statements and creates treeview items with filenames.
;--------------------------------------------------------------------
proc ScanFileForIncludes, .hFileName, .hTVParent
.tvi TVINSERTSTRUCT
.hthis dd ?
.ptr   dd ?
.cnt   dd ?
begin
        push    esi edi ebx

; Scan the file for includes.
        stdcall LoadBinaryFile, [.hFileName]
        test    eax, eax
        jz      .quit

        mov     [.ptr], eax
        mov     [.cnt], ecx
        test    ecx, ecx
        jz      .finish

; Add the file to the project tree.
        mov     eax, [.hTVParent]
        mov     [.tvi.hParent], eax
        mov     [.tvi.hInsertAfter], TVI_SORT

        stdcall DetectFileType, [.hFileName]
        test    eax, eax
        jz      .finish

        mov     ecx, [eax+TFileType.iFileIcon]
        mov     [.tvi.item.iImage], ecx
        mov     [.tvi.item.iSelectedImage], ecx

        stdcall CreateFStruct, [.hFileName], eax
        mov     [.tvi.item.lParam], eax
        mov     [.tvi.item.mask], TVIF_TEXT or TVIF_IMAGE or TVIF_PARAM or TVIF_SELECTEDIMAGE

        stdcall GetRelativePath, [.hFileName]
        push    eax
        stdcall StrPtr, eax
        mov     [.tvi.item.pszText], eax

        lea     eax, [.tvi]
        invoke  SendDlgItemMessageA, [hProjManager], ProjTreeCtrlID, TVM_INSERTITEM, 0, eax
        mov     [.hthis], eax

        stdcall StrDel ; from the stack

        mov     ecx, [.tvi.item.lParam]
        mov     [ecx+TOpenFile.projectID], eax

        mov     esi, [.ptr]
        mov     ecx, [.cnt]

.scan:
        mov     eax, [esi]
        cmp     al, ';'
        je      .comment
        or      eax, $20202020
        cmp     eax, 'incl'
        je      .maybe
.nextscan:
        inc     esi
        loop    .scan
.finish:
        stdcall FreeMem, [.ptr]
.quit:
        pop     ebx edi esi
        return

.comment:
        cmp     byte [esi], $0d
        je      .scan
        cmp     byte [esi], $0a
        je      .scan
        inc     esi
        loop    .comment
        jmp     .finish

.maybe:
        add     esi, 4
        sub     ecx, 4
        jbe     .finish

        mov     eax, [esi]
        and     eax, $ffffff
        or      eax, $202020
        cmp     eax, 'ude'
        je      .found
        add     esi, 3
        sub     ecx, 3
        jbe     .finish
        jmp     .scan

; Include statement
.found:
        add     esi, 3
        sub     ecx, 3
        jbe     .finish

.foundloop:
        mov     al, [esi]
        cmp     al, '"'
        je      .begin
        cmp     al, "'"
        je      .begin

        cmp     al, ' '
        je      .next
        cmp     al, $9
        jne     .nextscan

.next:
        inc     esi
        loop    .foundloop
        jmp     .finish

.begin:
        inc     esi
        dec     ecx
        jz      .finish
        mov     edi, esi
.scanend:
        cmp     [esi], al
        je      .nameok
        inc     esi
        loop    .scanend
        jmp     .finish
.nameok:
        push    ecx
        mov     byte [esi], 0
;        stdcall ConvertPath, edi
;        push    eax
;        stdcall StrPtr, eax
        stdcall ScanFileForIncludes, edi, [.hthis]
;        stdcall StrDel ; from the stack
        pop     ecx
        inc     esi
        dec     ecx
        jnz     .scan
        jmp     .finish
endp




proc EmptyVarList, .ptrVarArray
begin
        push    esi ebx

        mov     esi, [.ptrVarArray]
        test    esi, esi
        jz      .exit

        cmp     dword [esi], 0
        je      .array_clear

; now delete all strings in the array.
        mov     ebx, [esi]
        mov     ecx, [ebx+TArray.count]
        lea     ebx, [ebx+TArray.array]
        jecxz   .delok

.delloop:
        stdcall StrDel, [ebx+TEnvVarType.pName]
        stdcall StrDel, [ebx+TEnvVarType.pValue]

        add     ebx, sizeof.TEnvVarType
        loop    .delloop
.delok:
; so free the array itself.

        stdcall FreeMem, [esi]

.array_clear:
        stdcall CreateArray, sizeof.TEnvVarType
        mov     [esi], eax

.exit:
        pop     ebx esi
        return
endp






proc _AddKeyValueToArray, .ptrVarArray, .hValue, .hKey
begin
        push    esi

        mov     esi, [.ptrVarArray]
        mov     esi, [esi]
        mov     ecx, [esi+TArray.count]
        lea     esi, [esi+TArray.array-sizeof.TEnvVarType]
        inc     ecx

.searchloop:
        lea     esi, [esi+sizeof.TEnvVarType]
        dec     ecx
        jz      .notfound

        stdcall StrCompNoCase, [esi+TEnvVarType.pName], [.hKey]
        jnc     .searchloop

; found - append it.
        stdcall StrCharCat, [esi+TEnvVarType.pValue], '|'
        stdcall StrCat, [esi+TEnvVarType.pValue], [.hValue]
        stdcall   StrDel, [.hValue]

        pop     esi
        return

.notfound:
        mov     ecx, [.ptrVarArray]
        stdcall AddArrayItems, [ecx], 1
        mov     [ecx], edx

        mov     ecx, [.hValue]
        mov     edx, [.hKey]
        mov     [eax+TEnvVarType.pValue], ecx
        mov     [eax+TEnvVarType.pName], edx

        pop     esi
        return
endp



;********************************************************************************************************
; searches the project variables for  given name and copies the value (active one) in the buffer [.dest]
; Returns:
;   eax - count of copied chars.
;   CF = 1 - found and copied.
;   CF = 0 - not found.
;********************************************************************************************************

proc SearchForProjectVariable, .hName, .dest
begin
        push    esi edi ecx

        mov     esi, [poVarArray]
        test    esi, esi
        jz      .notfound

        mov     ecx, [esi+TArray.count]
        lea     esi, [esi+TArray.array]

        jecxz   .notfound
.loop:
        stdcall StrCompNoCase, [esi+TEnvVarType.pName], [.hName]
        jc      .found
        lea     esi, [esi+sizeof.TEnvVarType]
        loop    .loop

.notfound:
        clc
        xor     eax, eax
        pop     ecx edi esi
        return

.found:
        mov     edi, [.dest]
        stdcall StrPtr, [esi+TEnvVarType.pValue]
        mov     esi, eax
        xor     eax, eax

.copy:
        mov     cl, [esi]
        lea     esi, [esi+1]

        cmp     cl, 0
        je      .endofcopy

        cmp     cl, '|'
        je      .endofcopy

        mov     [edi], cl
        inc     eax
        lea     edi, [edi+1]
        jmp     .copy

.endofcopy:
        stc
        pop     ecx edi esi
        return
endp